% Demo: Level-shifted SCF for NEPv with single particle Hamiltonian. 
%
clear; close all;
set(0, 'DefaultLegendInterpreter', 'latex');
set(0,'defaultTextInterpreter','latex');
rng(1);

% problem size
n = 10;
k = 2;

% parameter
alpha = 1;
maxit = 350; % SCF
mtol = 1.0E-15;

% coefficient matrices
L = 2*diag(ones(n,1)) - diag(ones(n-1,1),1) - diag(ones(n-1,1),-1);
rhofun = @(V) diag(V*V');
H = @(V) L + alpha*diag(L\rhofun(V));

% initial vectors for SCF
[V0,D0] = eig(L);
V0 = V0(:,1:k);
D0 = D0(1:k,1:k);
V00 = V0; 


% -------------------------------------------------------------
%   First run of level-shifted SCF: to obtain `exact' solution.
% -------------------------------------------------------------

V0 = V00;   % initialize 
Hv = H(V0);
Resls = norm(Hv*V0-V0*D0);
sinv = 1/min(svd(L));
s_est = alpha*sinv*1.5+max(svd(L));  % shift, usually too pessimistic
for i = 1:10000
    [V,D] = eig(Hv - s_est*V0*V0');
    V0 = V(:,1:k);
    D0 = V0'*Hv*V0;
    Hv = H(V0);
    Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
    if(Resls(end)< mtol*n)
        break
    end
end
if Resls(end)> mtol*n
    disp('Level-shifted SCF not converge: increase maxit.')
    return
end


% -------------------------------------------------------------
% Plot spectral radius rho(L_sigma) as a function of sigma
% -------------------------------------------------------------

% 1. define operator L sigma: Op
[V,D] = eig(Hv);
lam = diag(D);
lam1 = lam(1:k); lam2 = lam(k+1:end);
Ss = @(s) 1./(lam2 - lam1'+s);
Vs = V(:,1:k); Vst = V(:,k+1:end);
Op = @(X,s) Ss(s).*( Vst'* (2*alpha*diag(L\diag(Vst*X*Vs'))) *  Vs) -s*Ss(s).*X;

%     define operator Qs and estimate mu_max
Opqs = @(X) 2*alpha*Vst'*diag(L\diag(Vst*X*Vs'))*Vs + diag(lam2)*X - X*diag(lam1);
Afun = @(x) reshape(Opqs(reshape(x, n-k, k)),[],1);
mux = eigs(Afun, (n-k)*k, 2, 'bothendsreal');
mu_max = max(mux);
mu_min = min(mux);
deltas = lam(k+1) - lam(k);
spans = lam(end) - lam(1);
rhobnd = @(s) max(abs([mu_max./(s+deltas)-1;  mu_min./(s+spans)-1]));
fs = @(s) mu_max./(s+deltas) + mu_min./(s+spans)-2;
s_opt = fzero(fs, [0,10]); % bnd optimal 
s_theoretic = mu_max/2-deltas; % theoretical optimal

% 2. compute spectral radius for various sigma
SIGMA2 = linspace(0, s_est*1.2, 300);
for ii = 1:length(SIGMA2)
    s = SIGMA2(ii);
    Afun = @(x) reshape(Op(reshape(x, n-k, k), s),[],1);
    lammax = eigs(Afun, (n-k)*k, 1);
    spLs2(ii) = abs(lammax);
end

% 3. plot 
figure(2);
plot(SIGMA2, spLs2, '-', 'linewidth', 2); hold on;
xlabel('$\sigma$')
ylabel('spectral radius')
plot([-.5, SIGMA2(end)], [1,1], '--k')
plot([s_est, s_est], [min(spLs2), max(spLs2)*1.05], '--k')
text(s_est*1.01, 0.4*(min(spLs2) + max(spLs2)), 'a priori', 'fontsize', 12)
plot([s_theoretic, s_theoretic], [min(spLs2), max(spLs2)*1.05], '--k')
xlim([-.5, SIGMA2(end)])
ylim([min(spLs2)*0.95, max(spLs2)*1.05])
[spls_opt, idx] = min(spLs2);
sigma_opt = SIGMA2(idx);


% -------------------------------------------------------------
%   SCF convergence for selected shifts 
% -------------------------------------------------------------

XVAR = []; YVAR =[]; STR =[];
SIGMA = logspace(log10(sigma_opt), log10(s_est), 5);
for ii = 1:length(SIGMA)
    s = SIGMA(ii);
    Afun = @(x) reshape(Op(reshape(x, n-k, k), s),[],1); % NOTE: Modify for complex input 
    lammax = eigs(Afun, (n-k)*k, 1);
    spLs(ii) = abs(lammax);

    % level-shifted SCF
    V0 = V00;
    Hv = H(V0);
    Resls = norm(Hv*V0-V0*D0);
    VVls = [];
    VVls{1} = V0;
    for i = 1:maxit
        [V,D] = eig(Hv - s*V0*V0');
        V0 = V(:,1:k);
        D0 = V0'*Hv*V0;
        Hv = H(V0);
        Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
        VVls{i} = V0;
        if(Resls(end)< mtol*n)
            break
        end
    end

    if abs(lammax) < 1
        figure(1); %convergence history
        semilogy(Resls, '-o'); hold on;
        xvar = min([250, length(Resls)-8]);
        XVAR = [XVAR; xvar];
        YVAR = [YVAR; Resls(xvar)*10];
        STR{ii} = ['$\sigma = $', num2str(s,'%4.2f')];
    end
end

% text label for sigma
text(XVAR, YVAR, STR, 'fontsize',12);
figure(1);
xlim([0,maxit])
ylim([1.0E-14, 1.0E1])
xlabel('SCF iteration');
ylabel('NEPv residual norm')

%figure(3);
%plot(SIGMA, spLs, '--o', 'linewidth', 2, 'markersize', 6); hold on;
%xlabel('$\sigma$')
%%ylabel(['$\rho(\mathcal L_\sigma)$'])
%ylabel('spectral radius')
%plot([0, SIGMA(end) + 2], [1,1], '--k')
%plot([s_est, s_est], [min(spLs), max(spLs)*1.05], '--k')
%text(s_est*1.02, 0.5*(min(spLs) + max(spLs)), '$\sigma$ bnd', 'fontsize', 20)
%xlim([0, SIGMA(end) + 2])
%ylim([min(spLs), max(spLs)*1.05])

return;
