% Demo: Plain SCF for NEPv with single particle Hamiltonian.
%
clear; close all;
set(0, 'DefaultLegendInterpreter', 'latex');
set(0,'defaultTextInterpreter','latex');
rng(1); 

% problem size
n = 10;
k = 2;

% range for alpha
a = 0.05;
b = 1.0; 
alpha = a:0.05:b;

% parameters
maxit = 10000;      % SCF
ALPHACONV = [0.5, 0.7, 0.75, 0.8, 0.85, 0.9]; % alpha values for convergence plot
rxlim = 400;        % plotting range for convergence
mtol = 1.0E-15;

% coefficient matrix H(V)
L = 2*diag(ones(n,1)) - diag(ones(n-1,1),1) - diag(ones(n-1,1),-1);
rho = @(V) diag(V*V');

[V0,D0] = eig(L);
V0 = V0(:,1:k);
D0 = D0(1:k,1:k);


% -------------------------------------
%   MAIN
% -------------------------------------

for nn = 1:length(alpha)
    H = @(V) L + alpha(nn)*diag(L\rho(V));
    V00 = V0;
    
    % ---------------------------------------- 
    %   plain SCF iteration
    % ---------------------------------------- 
    V0 = V00;
    Hv = H(V0);
    Res = norm(Hv*V0-V0*D0);
    VV = [];
    VV{1} = V0;
    for i = 1:maxit
        [V,D] = eig(Hv);
        V0 = V(:,1:k);
        D0 = D(1:k,1:k);
        Hv = H(V0);
        Res =[ Res, norm(Hv*V0-V0*D0)];
        VV{i} = V0;
        if(Res(end)< mtol*n)
            break
        end
    end

    RES{nn} = Res; % residual norm of soultion

    % plot the convergence history for selected alpha
    if find( abs(alpha(nn) - ALPHACONV) <= eps)
        figure(1); 
        semilogy(Res, '-o'); hold on;
        xvar = min([rxlim, length(Res)-10]);
        text(xvar, Res(xvar)*10, ['$\alpha =$ ', num2str(alpha(nn))], 'fontsize', 12);
    end
    Vhist{nn} = VV;

    % ---------------------------------------- 
    %   level-shifted SCF iteration, if needed
    %   (by default, we choose a large shift, which can be slow in convergence)
    % ---------------------------------------- 

    if Res(end) > mtol*n 
        V0 = V00;
        Hv = H(V0);
        Resls = norm(Hv*V0-V0*D0);
        xi = alpha(nn)*n; % level shift parameter (relative large, not optimal, can be slow)
        for i = 1:10000
            [V,D] = eig(Hv - xi*V0*V0');
            V0 = V(:,1:k);
            D0 = V0'*Hv*V0;
            Hv = H(V0);
            Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
            if(Resls(end)< mtol*n)
                VVls = V0;
                break
            end
        end
    end


    % ---------------------------------------- 
    %  convergence factors:
    % ---------------------------------------- 
    
    % 1. estimate observed convergence rate 
    %
    if Res(end) < 1.0E-8            % select sequence close to convergence 
        ns = find(Res< 1.0E-8, 1);  % index reaching 1.0E-8
        ne = find(Res< 1.0E-10, 1); % index reaching 1.0E-10
        dd = polyfit([ns:ne],log(Res(ns:ne)),1);
        Conv(nn) = exp(dd(1));      % plain SCF 
    end

    % 2. eta_infty (spectral radius)
    %
    if Res(end) < mtol*n % select 'exact' solution 
        V0 = VV{end};
    elseif Resls(end) < mtol*n 
        V0 = VVls;
    else 
        Resls(end) 
        disp('Level-shfifted SCF not converge: increase the tolerance or choose a larger shift.')
    end
    Hv = H(V0); 
    [V, D] = eig(Hv); 

    % 2.1 define linear operators
    lam = diag(D);
    lam1 = lam(1:k); lam2 = lam(k+1:end);
    Ss = 1./(lam2 - lam1');
    Vs = V(:,1:k); Vst = V(:,k+1:end);
    Op = @(X) 2*alpha(nn)* Ss.*( Vst'*diag(L\diag(Vst*X*Vs'))*Vs);
    Ops = @(Y) 2*alpha(nn)*( Vst'*diag(L'\diag(Vst*(Ss.*Y)*Vs'))*Vs);
    
    % 2.2 spectral radius by eigs
    Afun = @(x) reshape(Op(reshape(x, n-k, k)),[],1); % NOTE: Modify for complex input 
    lammax = eigs(Afun, (n-k)*k, 1);
    Convest_sp(nn) = lammax;

    % 3. eta ( operator 2 norm)
    %
    Afun = @(x) reshape(Ops(Op(reshape(x, n-k, k))), [],1); % NOTE: Modify for complex input 
    lammax = eigs(Afun, (n-k)*k, 1);
    Convest_2norm(nn) = sqrt(lammax);
    
    % 4. Cai 2018 
    %
    deltas = min(lam2) - max(lam1);
    TOp = @(X) 2*alpha(nn)*( Vst'*diag(L\diag(Vst*X*Vs'))*Vs);
    TOps = @(Y) 2*alpha(nn)*( Vst'*diag(L'\diag(Vst*(Y)*Vs'))*Vs);
    Afun = @(x) reshape(TOps(TOp(reshape(x, n-k, k))), [],1); % NOTE: Modify for complex input 
    lammax = eigs(Afun, (n-k)*k, 1);
    Convest_cai(nn) = sqrt(lammax)/deltas;
    
end

% plot convergence factors
figure(1); 
xlim([0,500])
ylim([1.0E-14, 1.0E1])
xlabel('SCF iteration');
ylabel('NEPv residual norm')

figure(2)
hh = plot(alpha(1:length(Conv)), Conv, '-ok', 'DisplayName', 'observed', 'linewidth', 1, 'Markersize', 6); hold on
h = hh;
hh = plot(alpha, Convest_cai, '--sb', 'DisplayName', '$\eta_{\rm czbl}$ (Cai et al., 2018) ', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];
hh = plot(alpha, Convest_2norm, '--xr', 'DisplayName', '$\eta_{\sup}$', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];
hh = plot(alpha, Convest_sp, '--+m', 'DisplayName', '$\eta_{\sup,\infty}$', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];

plot([0, alpha(end)], [1,1], '--k')

plot([0.6, 0.6], [0,2], '--b')
plot([0.85, 0.85], [0,2], '--r')

legend(h, 'location', 'northwest', 'Fontsize', 11);
xlim([min(alpha)*(0.9), max(alpha)])
ylim([0,max(Convest_cai)*1.1])
xlabel('$\alpha$')
ylabel('convergence factor')

% check convergence rate at the boundary 
idx = find(Convest_sp>1,1);
factors = [Conv(end); Convest_sp(idx-1); Convest_2norm(idx-1);
Convest_cai(idx-1)];
disp(['conv factors at alpha=', num2str(alpha(idx))])
factors'

