% Demo: Level-shifted SCF for NEPv in solving Gross–Pitaevskii equation
%

clear; close all;
set(0, 'DefaultLegendInterpreter', 'latex');
set(0,'defaultTextInterpreter','latex');
rng(1);

EXAMPLE = 1;    % 1: symmetric potential (x.^2+y.^2)/2;     
                % 2: non-symmetric potential (x.^2+100*y.^2)/2; 

if EXAMPLE == 1
    % problem size
    N = 10; % grid num 
    n = N*N;
    k = 1; 

    % parameter
    beta = 5.0;
    mtol = 1.0E-12;
    maxit = 300; % SCF

    Vf = @(x,y) (x.^2+y.^2)/2;      % potential function
else
    % problem size
    N = 10; % grid num 
    n = N*N;
    k = 1; 

    % parameter
    beta = 3.0;
    mtol = 1.0E-12;
    maxit = 300; % SCF
    
    Vf = @(x,y) (x.^2+100*y.^2)/2;  % potential function
end
    

% -------------------------------------
% GP discretization 
% -------------------------------------

L = 1;          % domain [-L, L]
h = 2*L/(N+1);  % stepsize
omega = 0.85;

DN = (diag(ones(N-1,1),1) - diag(ones(N-1,1),-1))/2/h; % central difference for 1st deriv
D2N = (diag(ones(N-1,1),1) -diag(ones(N,1))*2 +diag(ones(N-1,1),-1))/h/h; % for 2nd deriv
LN = kron(D2N,eye(N)) + kron(eye(N), D2N);
xx = linspace(-L, L, N+2); xx = xx(2:end-1);
LFN = kron(diag(xx), DN) - kron(DN, diag(xx));
[XX, YY] = meshgrid(xx,xx);
Vtilde = arrayfun(Vf, XX,YY);
Vtilde = Vtilde(:); % vectorize
A0tilde_real = -0.5*LN + diag(Vtilde);
A0tilde_imag = -1*omega*LFN;

A0 = h^2*(A0tilde_real + 1i*A0tilde_imag);

Bv = @(v) diag(abs(v).^2);

% coefficient matrix H(v)
H = @(v)  A0 + Bv(v)*beta;

% the derivative operators
Op0 = @(beta, Ss, Vst, Vs, X) 2*beta* Ss.*( Vst'*diag(real(conj(Vs).*(Vst*X)))*Vs);
Ops0 = @(beta, Ss, Vst, Vs, Y) 2*beta*(Vst'*real(diag(diag(Vst*(Ss.*Y)*Vs')))*Vs); 
 
% utility functions for eigs.
pp = n - k; NN = pp * k;
x2X = @(x) reshape(x(1:NN)+1i*x(NN+1:end), pp, []);
X2x = @(X) [real(X(:)); imag(X(:))];


% -------------------------------------------------------------
%   First run of level-shifted SCF: to obtain `exact' solution.
% -------------------------------------------------------------

[V0,D0] = eig(A0);
V0 = V0(:,1:k);
D0 = D0(1:k,1:k);
V00 = V0;
Hv = H(V0);
Resls = norm(Hv*V0-V0*D0);
xi = 1; % level-shift, typically ok, for fast simulation
for i = 1:10000
    [V,D] = eig(Hv - xi*V0*V0');
    V0 = V(:,1:k);
    D0 = V0'*Hv*V0;
    Hv = H(V0);
    Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
    VVls{i} = V0;
    if(Resls(end)< mtol*n)
        break
    end
end
if Resls(end)> mtol*n
    disp('Level-shifted SCF not converge: increase maxit.')
    return
end



% -------------------------------------------------------------
% Plot spectral radius rho(L_sigma) as a function of sigma
% -------------------------------------------------------------

% 1. define operator L sigma: Op
%
s_est = (norm(A0)+3*beta)/2; % to change...
[V,D] = eig(Hv);
lam = diag(D);
lam1 = lam(1:k); lam2 = lam(k+1:end);
Ss = @(s) 1./(lam2 - lam1'+s); % Ss = 1./(lam2 - lam1');
Vs = V(:,1:k); Vst = V(:,k+1:end);
Op = @(X,s) Op0(beta, Ss(s), Vst, Vs, X)-s*Ss(s).*X;

%     define operator Qs and estimate mu_max
Opqs = @(X) 2*beta*Vst'*diag(real(conj(Vs).*(Vst*X)))*Vs + diag(lam2)*X - X*diag(lam1);
Afun = @(x) X2x(Opqs(x2X(x)));
mux = eigs(Afun, 2*NN, 2, 'bothendsreal');
mu_max = max(mux);
mu_min = min(mux);
deltas = lam(k+1) - lam(k);
spans = lam(end) - lam(1);
rhobnd = @(s) max(abs([mu_max./(s+deltas)-1;  mu_min./(s+spans)-1]));
s_theoretic = mu_max/2-deltas;

% 2. compute spectral radius for various sigma
%
SIGMA2 = linspace(0, s_est*1.2, 300);
for ii = 1:length(SIGMA2)
    s = SIGMA2(ii);
    Lbar = @(x) X2x(Op(x2X(x),s));
    lammax = eigs(Lbar, 2*NN, 1);
    spLs2(ii) = abs(lammax);
end

% 3. plot
%
figure(2);
plot(SIGMA2, spLs2, '-', 'linewidth', 2); hold on;
xlabel('$\sigma$')
ylabel('spectral radius')
plot([-.5, SIGMA2(end)], [1,1], '--k')
plot([s_est, s_est], [min(spLs2), max(spLs2)*1.05], '--k')
text(s_est*1.01, 0.4*(min(spLs2) + max(spLs2)), 'a priori', 'fontsize', 12)
plot([s_theoretic, s_theoretic], [min(spLs2), max(spLs2)*1.05], '--k')
xlim([-.5, SIGMA2(end)])
ylim([min(spLs2)*0.95, max(spLs2)*1.05])


% -------------------------------------------------------------
%   SCF convergence for selected shifts 
% -------------------------------------------------------------

XVAR = []; YVAR =[]; STR =[];
[spls_opt, idx] = min(spLs2);
sigma_opt = SIGMA2(idx);
if EXAMPLE == 1
    SIGMA = [sigma_opt, logspace(log10(0.85), log10(s_est), 4)];
else % EXAMPLE 2
    SIGMA = [sigma_opt, logspace(log10(0.74), log10(s_est), 4)];
end
figure(1); %convergence history
for ii = 1:length(SIGMA)
    s = SIGMA(ii);
    V0 = V00;
    Hv = H(V0);
    Resls = norm(Hv*V0-V0*D0);
    VVls = [];

    % level-shifted SCF
    for i = 1:maxit
        [V,D] = eig(Hv - s*V0*V0');
        [~,idx] = sort(real(diag(D)),'ascend');
        V = V(:,idx);
        V0 = V(:,1:k);
        D0 = V0'*Hv*V0;
        Hv = H(V0);
        Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
        if(Resls(end)< 1.0E-14)
            break
        end
    end

    semilogy(Resls, '-o'); hold on;
    xvar = min([250, length(Resls)-10]);
    XVAR = [XVAR; xvar];
    YVAR = [YVAR; Resls(xvar)*10];
    STR{ii} = ['$\sigma = $', num2str(s,'%4.2f')];
end

% text label for sigma
text(XVAR, YVAR, STR, 'fontsize',12);
figure(1);
xlim([0,maxit])
ylim([1.0E-14, 1.0E0])
xlabel('SCF iteration');
ylabel('NEPv residual norm')


return
