% Demo: SCF for NEPv of Gross–Pitaevskii equation
%
clear; close all;
set(0, 'DefaultLegendInterpreter', 'latex');
set(0,'defaultTextInterpreter','latex');
rng(1);

EXAMPLE = 2;    % 1: symmetric potential (x.^2+y.^2)/2;     
                % 2: non-symmetric potential (x.^2+100*y.^2)/2; 

if EXAMPLE == 2
    % problem size
    N = 10; % grid num 
    n = N*N;
    k = 1; 

    % range for beta
    a = 0.5;
    b = 5.0; 
    beta = a:0.5:b;

    % parameter
    mtol = 1.0E-15;
    maxit = 400; % SCF
    BETACONV = [1.5, 3, 3.5, 4];    % beta values for convergence plot
    rxlim = 250;                    % plotting range for convergence

    Vf = @(x,y) (x.^2+y.^2)/2;      % potential function
else
    % problem size
    N = 10; % grid num 
    n = N*N;
    k = 1; 

    % range for beta
    a = 0.6;
    b = 3.0; 
    beta = a:0.4:b;

    % parameter
    mtol = 1.0E-15;
    maxit = 800; % SCF
    BETACONV = [1, 1.8, 2.2, 2.6];  % beta values for convergence plot
    rxlim = 250;                    % plotting range for convergence
    
    Vf = @(x,y) (x.^2+100*y.^2)/2;  % potential function
end
    

% -------------------------------------
% GP discretization 
% -------------------------------------

L = 1;          % domain [-L, L]
h = 2*L/(N+1);  % stepsize
omega = 0.85;

DN = (diag(ones(N-1,1),1) - diag(ones(N-1,1),-1))/2/h; % central difference for 1st deriv
D2N = (diag(ones(N-1,1),1) -diag(ones(N,1))*2 +diag(ones(N-1,1),-1))/h/h; % for 2nd deriv
LN = kron(D2N,eye(N)) + kron(eye(N), D2N);
xx = linspace(-L, L, N+2); xx = xx(2:end-1);
LFN = kron(diag(xx), DN) - kron(DN, diag(xx));
[XX, YY] = meshgrid(xx,xx);
Vtilde = arrayfun(Vf, XX,YY);
Vtilde = Vtilde(:); % vectorize
A0tilde_real = -0.5*LN + diag(Vtilde);
A0tilde_imag = -1*omega*LFN;
A0 = h^2*(A0tilde_real + 1i*A0tilde_imag);

Bv = @(v) diag(abs(v).^2);

% coefficient matrix H(v)
H0 = @(v, beta)  A0 + Bv(v)*beta;

% the derivative operators
Op0 = @(beta, Ss, Vst, Vs, X) 2*beta* Ss.*( Vst'*diag(real(conj(Vs).*(Vst*X)))*Vs);
Ops0 = @(beta, Ss, Vst, Vs, Y) 2*beta*(Vst'*real(diag(diag(Vst*(Ss.*Y)*Vs')))*Vs); 
 
% utility functions for eigs.
pp = n - k; NN = pp * k;
x2X = @(x) reshape(x(1:NN)+1i*x(NN+1:end), pp, []);
X2x = @(X) [real(X(:)); imag(X(:))];


% -------------------------------------
%   MAIN
% -------------------------------------

for nn = 1:length(beta)
    H = @(v) H0(v, beta(nn));
    rng(1);
    V00 = orth(rand(length(A0), k)+1i*rand(length(A0), k) -.5*(1+1i));
            % initialize with random vectors (eigvec of A0 converges too fast)


    % ---------------------------------------- 
    %   plain SCF iteration
    % ---------------------------------------- 
    V0 = V00;
    Hv = H(V0);
    D0 = V0'*Hv*V0;
    Res = norm(Hv*V0-V0*D0);
    VV = [];
    VV{1} = V0;
    for i = 1:maxit
        [V,D] = eig(Hv);
        V0 = V(:,1:k);
        D0 = D(1:k,1:k);
        Hv = H(V0);
        Res =[ Res, norm(Hv*V0-V0*D0)];
        VV{i} = V0;
        if(Res(end)< mtol*n)
            break
        end
    end

    RES{nn} = Res; % residual norm of soultion

    % plot the convergence history for selected alpha
    if find( abs(beta(nn) - BETACONV) <= eps)
        figure(1); 
        semilogy(Res, '-o'); hold on; drawnow;
        xvar = min([rxlim, length(Res)]);
        text(xvar, Res(xvar)*5, ['$\beta$ = ', num2str(beta(nn))], 'fontsize', 12);
    end
    Vhist{nn} = VV;

    
    % ---------------------------------------- 
    %   level-shifted SCF iteration, if needed
    %   (by default, we choose a large shift, which can be slow in convergence)
    % ---------------------------------------- 

    if Res(end) > mtol*n 
        V0 = V00;
        Hv = H(V0);
        Resls = norm(Hv*V0-V0*D0);
        xi = beta(nn); % level shift parameter (relative large, not optimal, can be slow)
        for i = 1:10000
            [V,D] = eig(Hv - xi*V0*V0');
            V0 = V(:,1:k);
            D0 = V0'*Hv*V0;
            Hv = H(V0);
            Resls =[ Resls, norm(Hv*V0-V0*D0)]; % res, w.r.t., original NEPv
            if(Resls(end)< mtol*n)
                VVls = V0;
                break
            end
        end
    end

    % ---------------------------------------- 
    %  convergence factors:
    % ---------------------------------------- 
    
    % 1. estimate observed convergence rate 
    %
    if Res(end) < 1.0E-8            % select sequence close to convergence 
        ns = find(Res< 1.0E-8, 1);  % index reaching 1.0E-8
        ne = find(Res< 1.0E-10, 1); % index reaching 1.0E-10
        dd = polyfit([ns:ne],log(Res(ns:ne)),1);
        Conv(nn) = exp(dd(1));      % plain SCF 
    end

    % 2. eta_infty (spectral radius)
    %
    if Res(end) < mtol*n % select 'exact' solution 
        V0 = VV{end};
    elseif Resls(end) < mtol*n 
        V0 = VVls;
    else 
        Resls(end) 
        disp('Level-shfifted SCF not converge: increase the tolerance or choose a larger shift.')
    end
    Hv = H(V0); 
    [V, D] = eig(Hv); 

    % 2.1 define linear operators
    lam = diag(D);
    lam1 = lam(1:k); lam2 = lam(k+1:end);
    Ss = 1./(lam2 - lam1');
    Vs = V(:,1:k); Vst = V(:,k+1:end);
    Op = @(X) Op0(beta(nn), Ss, Vst, Vs, X);
    Ops = @(Y) Ops0(beta(nn), Ss, Vst, Vs, Y);

   
    Afun = @(x) X2x(Op(x2X(x))); % NOTE: Modify for complex input 
    lammax = eigs(Afun, 2*NN, 1);
    Convest_sp(nn) = lammax;

    % 3. eta ( operator 2 norm)
    %
    Afun = @(x) X2x(Ops(Op((x2X(x))))); % NOTE: Modify for complex input 
    lammax = eigs(Afun, 2*NN, 1);
    Convest_2norm(nn) = sqrt(lammax);
    
    % 4. Cai 2018 
    %
    deltas = min(lam2) - max(lam1);
    TOp = @(X) Op0(beta(nn), ones(size(Vst,2),size(Vs,2)), Vst, Vs, X);
    TOps = @(Y) Ops0(beta(nn), ones(size(Vst,2),size(Vs,2)), Vst, Vs, Y);
    Afun = @(x) X2x(TOps(TOp(x2X(x))));
    lammax = eigs(Afun, 2*NN, 1);
    Convest_cai(nn) = sqrt(lammax)/deltas;  
end

% plot convergence factors
figure(1); 
xlim([0,300])
ylim([mtol*n, 1.0E1])
xlabel('SCF iteration');
ylabel('NEPv residual norm')

figure(2)
hh = plot(beta(1:length(Conv)), Conv, '-ok', 'DisplayName', 'observed', 'linewidth', 1, 'Markersize', 6); hold on
h = hh;
hh = plot(beta, Convest_cai, '--sb', 'DisplayName', '$\eta_{\rm czbl}$ (Cai et al., 2018) ', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];
hh = plot(beta, Convest_2norm, '--xr', 'DisplayName', '$\eta_{\sup}$', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];
hh = plot(beta, Convest_sp, '--+m', 'DisplayName', '$\eta_{\sup,\infty}$', 'linewidth', 1, 'Markersize', 9);
h = [h, hh];

plot([beta(1)-.05, beta(end)], [1,1], '--k')

legend(h, 'location', 'northwest', 'Fontsize', 10);
xlim([min(beta)*(0.9), max(beta)])
xlabel('$\beta$')
ylabel('convergence factor')

% check convergence rate at the boundary 
idx = find(Convest_sp>1,1);
factors = [Conv(end); Convest_sp(idx-1); Convest_2norm(idx-1);
Convest_cai(idx-1)];
disp(['conv factors at alpha=', num2str(beta(idx))])
factors'



